#!/bin/bash
# SPDX-License-Identifier: MPL-2.0 OR LGPL-3.0-or-later
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# == MPL-2.0 ==
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# Alternatively, this Source Code Form may also (at your option) be used
# under the terms of the GNU Lesser General Public License Version 3, as
# described below:
#
# == LGPL-3.0-or-later ==
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or (at
#  your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
#  for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

# cargo's --crate-type= argument was stabilised in Rust 1.64 but our MSRV is
# 1.63. This script emulates --crate-type= by temporarily modifying the
# crate-type option in Cargo.toml.

set -Eeuo pipefail

bail() {
	echo "$*" >&2
	exit 1
}

SRC_ROOT="$(readlink -f "$(dirname "${BASH_SOURCE[0]}")/..")"

# Only parse the first argument if it is --crate-type=foo. The rest of the
# arguments get passed as-is.
crate_types=("rlib" "staticlib" "cdylib")
case "${1:-}" in
	--crate-type=*)
		arg="${1#*=}"
		IFS=', ' read -r -a crate_types <<<"$arg"
		shift 1
		;;
esac

[ "$#" -gt 0 ] || bail "usage: with-crate-type.sh [--crate-type=...] <command> [...]"

set -x

backup="$(mktemp "$SRC_ROOT/Cargo.toml.XXXXXX")"
sed -i".${backup##*.}" \
	"/^crate-type/ s/=.*/= [$(printf '"%s",' "${crate_types[@]}")]/" \
	"$SRC_ROOT/Cargo.toml"
# shellcheck disable=SC2064 # We want to expand the variables immediately.
trap "mv '$backup' '$SRC_ROOT/Cargo.toml'" EXIT

"$@"
